'' This is the initialization (main) routine for the Propeller "Operating System"
'' This set of objects is intended to provide a basic set of I/O including VGA
'' text video, a PS/2 keyboard, and management of one or more I2C memory busses,
'' each containing up to 256K of EEPROM.  One is always the boot I2C bus on I/O
'' pins 28 and 29.  Others can be added on any pair of I/O pins.  The VGA pins
'' and keyboard pins are those used by the Propeller Demo Board, Rev C.  Included
'' is a SPIN loader that can load a normal SPIN program from one of the EEPROMs
'' known to the system, from any location in the EEPROM.  The size of the program
'' is stored with the program by the IDE, but this can be overridden.
'' The I/O drivers are intended to run continuously in 3 COGs (2 for video and 1
'' for the keyboard).  Their work areas (including buffers) are kept at the end
'' of RAM memory so they should not be overwritten when a new program is loaded.
'' The new program, if it uses these video text, keyboard, and loader objects,
'' should be able to seamlessly use the video and keyboard and transfer control
'' further to other programs and back to this main routine for further interaction.
'' The loader will stop any COGs not known to it prior to loading a new progam.
'' Programs can register COGs intended to "persist" between programs.  These
'' must be running assembly language and use a work area that won't be overlaid.
'' There are methods provided by the loader to do simple work area allocation.

'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are copyrighted by Parallax, some with modifications by
'' others.  See the notes at the beginning of each source file for specifics.

CON
  _clkmode    = xtal1 + pll16x
  _xinfreq    = 5_000_000

  vgaPinGroup = 16                             ' Base pin for VGA pin group
  ps2KeyData  = 26                             ' Data pin for PS/2 keyboard
  ps2KeyClock = 27                             ' Clock pin for PS/2 keyboard
  i2cBootSCL  = 28                             ' I2C Clock pin for boot EEPROM
  i2cBootSDA  = 29                             ' I2C Data pin for boot EEPROM
  eepromDevice = %1010_0000                    ' Device address of EEPROM
  
  uniqueMark  = $965A3EC1                      ' Indicates that RAM was initialized

OBJ
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  vga     : "OS_vgaHiresText"                  ' Hires VGA driver modified for OS use
  vgaInit : "OS_vgaHiresTextInit"              '  Initialization and assembly portion
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  keyInit : "OS_keyboardInit"                  '  Initialization and assembly portion
  boot    : "OS_i2cObject"                     ' I2C driver used for "internal" boot EEPROM
  data    : "OS_i2cObject"                     ' I2C driver used for external data EEPROM
  debug   : "OS_debug"                         ' Debugging display routines

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed
                                               
  #1, processProbe, processInfo, processEdit   ' For command post-processing

'' Internal SPIN information at fixed locations in RAM/EEPROM image

  pbase      = $0006                           ' must be $0010
  vbase      = $0008                           ' number of longs loaded times 4
  dbase      = $000A                           ' above where $FFF9FFFF's get placed
  pcurr      = $000C                           ' points to Spin code
  dcurr      = $000E                           ' points to local stack
  
PUB initialize | t, y, i, j, z
  if long[ldr#initMarker] <> uniqueMark                    ' Check for a valid initialization
    repeat i from 0 to 7
      if i <> cogid                                        ' Stop all cogs except this one
        cogstop(i)
    long[ldr#memPtr]    := ldr#endFree                     ' Start at end of HUB memory
    t := ldr.allocate(ldr#systemSize)                      ' Allocate system work area
    long[ldr#systemPtr] := t
    long[t+ldr#sys_cog] := 1 << cogid                      ' Mark current cog in use
    long[t+ldr#sys_random] := ldr#initMarker               ' Initialize random seed
    wordfill(t+ldr#sys_eeproms,0,ldr#maxEEPROMs)           ' Clear EEPROMs bitmap table
    long[ldr#videoPtr]  := ldr.allocate(ldr#videoSize)     ' Allocate work area for video
    long[ldr#screenPtr] := ldr.allocate(ldr#screenSize)    ' Allocate screen buffer
    long[ldr#colorPtr]  := ldr.allocate(ldr#colorSize)     ' Allocate screen color buffer
    long[ldr#cursorPtr] := ldr.allocate(ldr#cursorSize)    ' Allocate cursor control block
    if not vgaInit.start(vgaPinGroup)                      ' Now start up the VGA driver
      abort false
    vga.str(string("Propeller Operating System - Version 1.013.012",vga#Cr))
    long[ldr#keybdPtr]  := ldr.allocate(ldr#keybdSize)     ' Allocate work area for keyboard
    if not keyInit.start(ps2KeyData,ps2KeyClock)           ' Now start up the keyboard driver
      abort false
    long[ldr#initMarker] := uniqueMark                     ' Basic initialization complete
    waitcnt(clkfreq * 2 + cnt)                             ' Wait for keyboard to initialize
  if not key.present
    vga.str(string(vga#Cr,"--> keyboard not present, can't continue"))
    abort long[ldr#initMarker] := 0
  boot.init(i2cBootSDA,i2cBootSCL)                         ' Initialize I2C driver for boot
  if not boot.devicePresent(eepromDevice)
    vga.str(string(vga#Cr,"--> boot eeprom not accessable, can't continue"))
    abort long[ldr#initMarker] := 0
  if (t := probeEEPROMs(i2cBootSCL)) == 0                  ' Probe boot eeprom for size
    vga.str(string(vga#Cr,"--> boot eeprom not readable/writable, can't continue"))
    abort long[ldr#initMarker] := 0
  word[long[ldr#systemPtr]+ldr#sys_eeproms+i2cBootSCL] := t
  vga.out(vga#Cr)
  repeat
    y := vga.windowRow
    if (t := \enterDoCommand(y)) < 0
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      vga.str(string("*** "))
      vga.str(-t)
      vga.out(vga#Cr)
    else
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      case t
        processProbe:
          repeat i from 0 to 30 step 2
            z~
            y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
            if y <> 0
              repeat j from 0 to 15
                if (1 << j & y) <> 0
                  z += 32
              vga.str(string("    Pin "))
              vga.decFixed(i,2)
              vga.str(string("  "))
              vga.decFixed(z,3)
              vga.str(string("K",vga#Cr))
        processInfo:
          z := boot.readLocation(eepromDevice,pbase,16,8)
          z := boot.readLocation(eepromDevice,pbase+1,16,8) << 8 | z
          vga.str(string("    boot EEPROM pbase = "))
          vga.hex(z,4)
          z := boot.readLocation(eepromDevice,vbase,16,8)
          z := boot.readLocation(eepromDevice,vbase+1,16,8) << 8 | z
          vga.str(string(", vbase = "))
          vga.hex(z,4)
          z := boot.readLocation(eepromDevice,dbase,16,8)
          z := boot.readLocation(eepromDevice,dbase+1,16,8) << 8 | z
          vga.str(string(", dbase = "))
          vga.hex(z,4)
          z := boot.readLocation(eepromDevice,pcurr,16,8)
          z := boot.readLocation(eepromDevice,pcurr+1,16,8) << 8 | z
          vga.str(string(", pcurr = "))
          vga.hex(z,4)
          z := boot.readLocation(eepromDevice,dcurr,16,8)
          z := boot.readLocation(eepromDevice,dcurr+1,16,8) << 8 | z
          vga.str(string(", dcurr = "))
          vga.hex(z,4)
          vga.out(vga#Cr)
        processEdit:
          editWindow
        other:
          vga.str(string("    result = "))
          vga.hex(t,4)
          vga.out(vga#Cr)

PUB editWindow | t
  vga.out6(vga#SetWin,1,5,1,vga#cols-5,vga#rows-2)
  vga.out6(vga#SetWin,2,0,1,5,vga#rows-2)
  vga.out3(vga#UseWin,2,vga#ClrScr)
  repeat t from 1 to vga#rows - 2
    vga.out3(vga#MoveXY,0,t-1)
    vga.decFixed(t,4)
  vga.out5(vga#UseWin,1,vga#Cursor,%01111,vga#ClrScr)
  repeat
    t := vga.windowEditor
  until t == $CC
  vga.out5(vga#UseWin,0,vga#Cursor,%00000,vga#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  vga.out6(vga#SetWin,1,1,y,vga#cols-1,1)
  vga.out2(vga#UseWin,0)
  vga.out4(vga#MoveXY,0,y,">")
  vga.out2(vga#UseWin,1)
  vga.out2(vga#Cursor,%01111)
  repeat
    t := vga.windowEditor
  until t == vga#Cr
  x~
  if (x := \getTextWord(x,@a,4,true," ")) < 0
    abort -string("no command given")
  case a
    $45564153:  ' save
      if (x := \getTextWord(x,@a,12,true," ")) < 0
        abort -string("missing file name")
    $424F5250:  ' probe
      if (x := \getNumber(x,@a)) < 0
        return processProbe
      if a < 32 and a & 1 == 0
        if (t := probeEEPROMs(a)) == 0
          abort -string("no eeprom(s) found on that pin pair")
        else
          word[long[ldr#systemPtr]+ldr#sys_eeproms+a] := t
          return processProbe
      else
        abort -string("pin number odd or not between 0 and 30")
    $204E5552:  ' run
      if (x := \getTextWord(x,@a,12,true," ")) < 0
        abort -string("missing file name")
    $54494445:  ' edit
      return processEdit
    $4F464E49:  ' info
      return processInfo
    other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if vga.windowAtEnd(x,0)
      abort -string("looking for non-blank at end of line")
  until vga.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := vga.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
  x := skipBlanks(x)
  c := vga.windowChar(x++,0)
  if (negative := c == "-") or c == "+"
    c := vga.windowChar(x++,0)
  if c < "0" or c > "9"
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := vga.windowChar(x++,0)
  repeat until c < "0" or c > "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := vga.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI probeEEPROMs(SCLpin) : bitmap | device, devAddr, addr, b0, b0a, b1, b1a, val, seed, t
'' This routine first attempts to see if an EEPROM is present.
'' A series of 32Kx8 address spaces could be provided by one
'' device or several.  This routine builds a bitmap showing which
'' have been provided if any and returns a value suitable for
'' placing in the system's EEPROM table.
  bitmap := 0
  seed := long[ldr#systemPtr] + ldr#sys_random
  data.init(SCLpin+1,SCLpin)
  if data.testBus == data#_none       ' Is there at least a pullup on SDA?
    repeat device from 0 to 7         ' If so, go through the EEPROM devices
      if data.devicePresent(eepromDevice | (device << 1))
        bitmap |= 1 << (device * 2)   ' Add the ones that respond to the bitmap
    if bitmap <> 0                    ' If any responded, we have to see which
      repeat device from 0 to 7       ' devices are 32Kx8 and which are 64Kx8
        if bitmap & (1 << (device * 2)) <> 0
          devAddr := eepromDevice | (device << 1)
          addr := long[seed]? & $7FFF ' Generate random 32K address
          b0 := data.readLocation(devAddr,addr,16,8)
          b1 := data.readLocation(devAddr,addr|$8000,16,8)
          repeat                      ' Generate random data not currently in
            val := long[seed]? & $FF  ' the first 32Kx8 address space or in the
          until val <> b0 and val <> b1  ' the second address space (if present)
          data.writeLocation(devAddr,addr|$8000,val,16,8)  ' Write the data presumably
          waitcnt(clkfreq / 200 + cnt)
          b0a := data.readLocation(devAddr,addr,16,8)      ' to the second address space
          b1a := data.readLocation(devAddr,addr|$8000,16,8)  ' and see where it appears
          if b0 == b0a and b1a == val  ' If the change is in the second 32K only
            bitmap |= 1 << (device * 2 + 1)  ' then we really have memory there
            data.writeLocation(devAddr,addr|$8000,b1,16,8)  ' Restore the original data
          else
            data.writeLocation(devAddr,addr,b0,16,8)  ' If not there, restore first 32K
          waitcnt(clkfreq / 200 + cnt)

'' Change log
'' 1.013   Use CONs in OS_keyboard for references in OS_keyboardInit
''         Corrections to assembly and SPIN OS_loader routines
''         Add table for attached EEPROMs and routine to probe for size
''         Added vga.outN routines for easier command output.
''         Added vga absolute screen and windowed access routines
''         Added control codes for scroll up and scroll down
''         Added simple commands for probing EEPROMs, boot EEPROM info
''         Always probe the boot EEPROM for size & record in table